 ! Informatik 1: Inst. f. Wissenschaftliches Rechnen, TU Dresden !
 !               Prof. Dr. Wolfgang V. Walter                    !
MODULE  POLYMOD              !!! Grundarithmetik fuer Polynome !!!
  IMPLICIT NONE ! keine Standardtypisierung => alles vereinbaren !
  PRIVATE       ! kein automatischer Export => PUBLIC verwenden  !
  PUBLIC :: GET, PUT, &                            ! Exportliste !
 &          OPERATOR(.ADD.), OPERATOR(.SUB.), OPERATOR(.MUL.),   &
 &          OPERATOR(.IGRAL.), OPERATOR(.DERIV.), OPERATOR(.AT.) 

  INTERFACE GET
    MODULE PROCEDURE  get
  END INTERFACE

  INTERFACE PUT
    MODULE PROCEDURE  put
  END INTERFACE

  INTERFACE OPERATOR (.ADD.)
    MODULE PROCEDURE  add
  END INTERFACE

  INTERFACE OPERATOR (.SUB.)
    MODULE PROCEDURE  sub
  END INTERFACE

  INTERFACE OPERATOR (.MUL.)
    MODULE PROCEDURE  mul
  END INTERFACE

  INTERFACE OPERATOR (.IGRAL.)
    MODULE PROCEDURE  integral
  END INTERFACE

  INTERFACE OPERATOR (.DERIV.)
    MODULE PROCEDURE  ableitung
  END INTERFACE

  INTERFACE OPERATOR (.AT.)
    MODULE PROCEDURE  horner
  END INTERFACE

CONTAINS 

! ------------------- !
!  CALL GET ( Poly )  !
! ------------------- !
  SUBROUTINE  get (p)
    REAL, DIMENSION( 0 : ), INTENT(OUT) ::  p
    WRITE(*,*) ' Bitte Koeffizienten eines Polynoms vom Grad ', &
   &   UBOUND(p,1), ' in aufsteigender Reihenfolge eingeben: '
    READ (*,*) p
  END SUBROUTINE  get

! ---------------------------- !
!  CALL PUT ( Poly [, Text] )  !
! ---------------------------- !
  SUBROUTINE  put (p, text)
    REAL, DIMENSION( 0 : ), INTENT(IN) ::  p
    CHARACTER ( LEN = * ), OPTIONAL    :: text
    INTEGER                            :: i
    IF (PRESENT(text))  WRITE(*, FMT='(A)', ADVANCE='NO') text
    WRITE(*,*) p(0), ( ' + ', p(i), '*x^', i, i=1, UBOUND(p,1) )
  END SUBROUTINE  put

! --------------------------- !
!  Poly .ADD. Poly  ->  Poly  !
! --------------------------- !
  FUNCTION  add (p, q)
    REAL, DIMENSION( 0 : ), INTENT(IN) ::  p, q
    REAL, DIMENSION( 0 : MAX(UBOUND(p,1), UBOUND(q,1)) ) ::  add
    INTEGER                            ::  degp, degq, mini, maxi
    degp = UBOUND(p,1) ;   degq = UBOUND(q,1)
    mini = MIN(degp, degq) ;   maxi = MAX(degp, degq)
    add(0:mini) = p(0:mini) + q(0:mini)
    IF ( degp < degq ) THEN
      add(mini+1:maxi) = q(mini+1:maxi)
    ELSE IF ( degp > degq ) THEN
      add(mini+1:maxi) = p(mini+1:maxi)
    END IF
  END FUNCTION  add

! --------------------------- !
!  Poly .SUB. Poly  ->  Poly  !
! --------------------------- !
  FUNCTION  sub (p, q)
    REAL, DIMENSION( 0 : ), INTENT(IN) ::  p, q
    REAL, DIMENSION( 0 : MAX(UBOUND(p,1), UBOUND(q,1)) ) ::  sub
    INTEGER                            ::  degp, degq, mini, maxi
    degp = UBOUND(p,1) ;   degq = UBOUND(q,1)
    mini = MIN(degp, degq) ;   maxi = MAX(degp, degq)
    sub(0:mini) = p(0:mini) - q(0:mini)
    IF ( degp < degq ) THEN
      sub(mini+1:maxi) = - q(mini+1:maxi)
    ELSE IF ( degp > degq ) THEN
      sub(mini+1:maxi) = p(mini+1:maxi)
    END IF
  END FUNCTION  sub

! --------------------------- !
!  Poly .MUL. Poly  ->  Poly  !
! --------------------------- !
  FUNCTION  mul (p, q)
    REAL, DIMENSION( 0 : ), INTENT(IN) ::  p, q
    REAL, DIMENSION( 0 : UBOUND(p,1) + UBOUND(q,1) ) ::  mul
    INTEGER                            ::  degp, degq, k
    degp = UBOUND(p,1) ;   degq = UBOUND(q,1)
    DO k = 0, degp + degq
      mul(k) = DOT_PRODUCT(p(MAX(0,k-degq):MIN(k,degp)),     &
     &                     q(MIN(k,degq):MAX(0,k-degp):-1))
    END DO
  END FUNCTION  mul

! ------------------------ !
!  .IGRAL. Poly  ->  Poly  !
! ------------------------ !
  FUNCTION  integral (p)
    REAL, DIMENSION( 0 : ), INTENT(IN)   ::  p
    REAL, DIMENSION( 0 : UBOUND(p,1)+1 ) ::  integral
    INTEGER                              ::  degp, k
    degp = UBOUND(p,1)
    integral(0) = 0.0
    DO k = 1, degp+1
      integral(k) = p(k-1) / k
    END DO
  END FUNCTION  integral

! ------------------------ !
!  .DERIV. Poly  ->  Poly  !
! ------------------------ !
  FUNCTION  ableitung (p)
    REAL, DIMENSION( 0 : ), INTENT(IN)   ::  p
    REAL, DIMENSION( 0 : UBOUND(p,1)-1 ) ::  ableitung
    INTEGER                              ::  degp, k
    degp = UBOUND(p,1)
    DO k = 0, degp-1
      ableitung(k) = (k+1) * p(k+1)
    END DO
  END FUNCTION  ableitung

! -------------------------- !
!  Poly .AT. REAL  ->  REAL  !
! -------------------------- !
  REAL FUNCTION  horner (p, a)
    REAL, DIMENSION( 0 : ), INTENT(IN) ::  p
    REAL, INTENT(IN)                   ::  a
    INTEGER                            ::  degp, k
    degp = UBOUND(p,1)
    horner = p(degp)
    DO k = degp-1, 0, -1
      horner = horner * a + p(k)
    END DO
  END FUNCTION  horner

END MODULE  POLYMOD




PROGRAM  POLY                !!! Testprogramm fuer Polynome !!!
  USE POLYMOD
  IMPLICIT NONE
  INTEGER                         ::  deg, k
  REAL                            ::  x
  REAL, DIMENSION(:), ALLOCATABLE ::  P, PP, PPP, Q

  WRITE(*,*) ' Bitte Grad der Polynome eingeben: '
  READ (*,*) deg
  ALLOCATE ( P(0:deg), Q(0:deg) )
  WRITE(*,*) ' Bitte Koeffizienten des Polynoms P eingeben: '
  CALL GET(P)
  WRITE(*,*) ' Bitte Koeffizienten des Polynoms Q eingeben: '
  CALL GET(Q)
  WRITE(*,*) ' Bitte Auswertestelle x eingeben: '
  READ (*,*) x
  CALL PUT(P, ' Polynom P: ')
  CALL PUT(.DERIV.(.IGRAL. P), ' Abl(Igr(P)): ')
  CALL PUT(.IGRAL.(.DERIV. P), ' Igr(Abl(P)): ')
  ALLOCATE ( PP(0:deg) )
  PP = P
  DO k = 1, deg
    ALLOCATE ( PPP(0:deg-k) )
    PPP = .DERIV. PP
    CALL PUT(PPP, ' Abl(...(P)...): ')
    DEALLOCATE ( PP )
    ALLOCATE ( PP(0:deg-k) )
    PP = PPP
    DEALLOCATE ( PPP )
  END DO
  CALL PUT(P, ' Polynom P: ')
  CALL PUT(.IGRAL. P, ' Igr(P): ')
  WRITE (*,*) ' P(', x, ')=', P .AT. x
  WRITE (*,*) ' Igr(P)(', x, ')=', .IGRAL. P .AT. x
  CALL PUT(Q, ' Polynom Q: ')
  CALL PUT(.IGRAL. Q, ' Igr(Q): ')
  WRITE (*,*) ' Q(', x, ')=', Q .AT. x
  WRITE (*,*) ' Igr(Q)(', x, ')=', .IGRAL. Q .AT. x
  CALL PUT(P .ADD. Q, ' P + Q: ') 
  CALL PUT(P .SUB. Q, ' P - Q: ') 
  CALL PUT(P .MUL. Q, ' P * Q: ') 
  CALL PUT(.IGRAL. P .MUL. .DERIV. Q, ' Igr(P) * Abl(Q): ') 
  CALL PUT(((P .ADD. Q) .SUB. P) .SUB. Q, ' P + Q - P - Q: ') 
END PROGRAM  POLY

